////////////////////////////////////////////////////////////////////////////////
//
//  R E L I E F . F R A G
//
//  Author: Dave Rodgers
//  Date:   Mar 11, 2006
//
//

varying vec3 eye;
varying vec3 texspaceEye;

varying vec3 normal;
varying vec3 tangent;
varying vec3 binormal;

varying vec2 texCoord;

uniform sampler2D bzTexture0; // RGB + specular
uniform sampler2D bzTexture1; // normal + height


uniform float depthFactor;


float ray_intersect(sampler2D reliefMap, vec2 dp, vec2 ds)
{
  const int linear_search_steps = 10; // orig = 20
  
  float size = 1.0 / float(linear_search_steps);
  float depth = 0.0;
  float bestDepth = 1.0;

  for (int i = 0; i < linear_search_steps - 1; ++i) {
    depth += size;
    float t = 1 - texture2D(reliefMap, dp + ds * depth).a;
    if (bestDepth > 0.996) {
      if (depth >= t) {
        bestDepth = depth;
      }
    }
  }
  depth = bestDepth;

  const int binary_search_steps = 5; // orig = 5

  for (int i = 0; i < binary_search_steps; ++i) {
    size *= 0.5;
    float t = 1 - texture2D(reliefMap, dp + ds * depth).a;
    if (depth >= t) {
      bestDepth = depth;
      depth -= 2.0 * size;
    }
    depth += size;
  }

  return (bestDepth);
}


void main()
{
  const float shininess = gl_FrontMaterial.shininess;
  const vec3 matDiff = gl_FrontMaterial.diffuse.rgb;
  const vec3 matSpec = gl_FrontMaterial.specular.rgb;

  const vec3 E = normalize(eye);
  float a = dot(normal, E);
  vec2 s = vec2(dot(-E, tangent), dot(-E, binormal));
  s *= (depthFactor / a);
  vec2 ds = s;
  vec2 dp = texCoord;
  float d = ray_intersect(bzTexture1, dp, ds);

  texCoord = dp + (ds * d);

  vec3 N = texture2D(bzTexture1, texCoord).xyz * 2.0 - 1.0;
  vec4 diffuse = texture2D(bzTexture0, texCoord);

  N.z = sqrt(1.0 - dot(N.xy, N.xy));
  N = normalize(N.x * tangent + N.y * binormal + N.z * normal);


  // Tally the main light
  vec3 Idiff = vec3(0.0);
  vec3 Ispec = vec3(0.0);

#ifdef LAST_LIGHT
  const vec3 Lm = normalize(-gl_LightSource[0].position.xyz);
  const vec3 Rm = -reflect(Lm, N);
  Idiff = gl_LightSource[0].diffuse.rgb * max(dot(N, Lm), 0.0);
  Ispec = gl_LightSource[0].specular.rgb * pow(max(dot(Rm, E), 0.0), shininess);

  // the point lights
  for (int i = 1; i <= LAST_LIGHT; i++) {
    if (gl_LightSource[i].ambient.a != 0.0) {
      const vec3 light = (gl_LightSource[i].position.xyz + eye);
      const float dist = length(light);
      const vec3 L = light / dist; // normalize
      const vec3 R = -reflect(L, N);

      const float atten =
				1.0 / (gl_LightSource[i].constantAttenuation +
               (gl_LightSource[i].linearAttenuation * dist) +
               (gl_LightSource[i].quadraticAttenuation * dist * dist));

      // diffuse
      const vec3 Tdiff = gl_LightSource[i].diffuse.rgb * max(dot(N, L), 0.0);
      Idiff = Idiff + (atten * Tdiff);

      // specular
      const vec3 Tspec = gl_LightSource[i].specular.rgb * pow(max(dot(R, E), 0.0), shininess);
      Ispec = Ispec + (atten * Tspec);
		}
  }
#endif // LAST_LIGHT  
  
  const vec3 sceneColor = gl_FrontLightModelProduct.sceneColor.rgb;

	
  gl_FragColor.a   = gl_FrontMaterial.diffuse.a;
  gl_FragColor.rgb = (diffuse.a * Ispec * matSpec) +
                     (diffuse.rgb * (sceneColor + (Idiff * matDiff)));
}
