--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--
--  file:     primitives.lua
--  author:   Dave Rodgers (aka: trepan)
--  date:     Jan 07, 2008
--  license:  LGPL 2.1
--  desc:     world primitve generator for bzflag
--
--
--  TODO:
--  - WorldText
--  - MeshDrawInfo
--  - fix pyramid flipz
--  - finish teleporter txcds (border and links)
--  
--
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

-- Automatically generated local definitions

local GL_QUADS              = GL.QUADS
local GL_TRIANGLES          = GL.TRIANGLES
local GL_TRIANGLE_FAN       = GL.TRIANGLE_FAN
local GL_TRIANGLE_STRIP     = GL.TRIANGLE_STRIP
local bzGetObstacleBorder   = bz.GetObstacleBorder
local bzGetObstacleFlipZ    = bz.GetObstacleFlipZ
local bzGetObstaclePosition = bz.GetObstaclePosition
local bzGetObstacleRotation = bz.GetObstacleRotation
local bzGetObstacleSize     = bz.GetObstacleSize
local bzGetObstacleTeam     = bz.GetObstacleTeam
local bzGetObstacleType     = bz.GetObstacleType
local glBeginEnd            = gl.BeginEnd
local glMultiTexCoord       = gl.MultiTexCoord
local glNormal              = gl.Normal
local glTexCoord            = gl.TexCoord
local glVertex              = gl.Vertex
local math_cos              = math.cos
local math_sin              = math.sin

local length                = math.vector.length
local normalize             = math.vector.normalize


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function GetDrawFuncs(verts, norms, txcds, tangs, binms)
  local V  = function(i) glVertex(verts[i])   end
  local N  = function(i) glNormal(norms[i])   end
  local TC = function(i) glTexCoord(txcds[i]) end
  local TN = function(i) glMultiTexCoord(2, tangs[i]) end 
  local BN = function(i) glMultiTexCoord(3, binms[i]) end
  return V, N, TC, TN, BN
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function DrawBoxElement(self, elementID)
  if (elementID == 1) then -- sides
    glBeginEnd(GL_QUADS, function(box)
      local V, N, TC, TN, BN = GetDrawFuncs(box.verts, box.norms, box.txcds,
                                            box.tangs, box.binms)
      N(1); TN(1); BN(1);
      TC(1); V(4); TC(2); V(1); TC(3); V(5); TC(4); V(8);
      N(2); TN(2); BN(2);
      TC(1); V(2); TC(2); V(3); TC(3); V(7); TC(4); V(6);
      N(3); TN(3); BN(3);
      TC(5); V(1); TC(6); V(2); TC(7); V(6); TC(8); V(5);
      N(4); TN(4); BN(4);
      TC(5); V(3); TC(6); V(4); TC(7); V(8); TC(8); V(7);
    end, self)
  elseif (elementID == 2) then -- botton and top
    glBeginEnd(GL_QUADS, function(box)
      local V, N, TC, TN, BN = GetDrawFuncs(box.verts, box.norms, box.txcds,
                                            box.tangs, box.binms)
      N(5); TN(5); BN(5);
      TC(9); V(4); TC(10); V(3); TC(11); V(2); TC(12); V(1);
      N(6); TN(6); BN(6);
      TC(9); V(5); TC(10); V(6); TC(11); V(7); TC(12); V(8);
    end, self)
  else
    error('DrawBoxElement called with elementID = ' .. tostring(elementID))
  end
end


local function DrawPyrElement(self, elementID)
  if (elementID == 1) then -- sides
    glBeginEnd(GL_TRIANGLES, function(pyr)
      local V, N, TC, TN, BN = GetDrawFuncs(pyr.verts, pyr.norms, pyr.txcds,
                                            pyr.tangs, pyr.binms)

      N(1); TN(1); BN(1);
      TC(1); V(4); TC(2); V(1); TC(3); V(5);
      N(2); TN(2); BN(2);
      TC(1); V(2); TC(2); V(3); TC(3); V(5);
      N(3); TN(3); BN(3);
      TC(1); V(1); TC(4); V(2); TC(5); V(5);
      N(4); TN(4); BN(4);
      TC(1); V(3); TC(4); V(4); TC(5); V(5);
    end, self)
  elseif (elementID == 2) then -- bottom
    glBeginEnd(GL_QUADS, function(pyr)
      local V, N, TC, TN, BN = GetDrawFuncs(pyr.verts, pyr.norms, pyr.txcds,
                                            pyr.tangs, pyr.binms)
      N(5); TN(5); BN(5);
      TC(6); V(4); TC(7); V(3); TC(8); V(2); TC(9); V(1);
    end, self)
  else
    error('DrawPyrElement called with elementID = ' .. tostring(elementID))
  end
end


local function DrawTeleElement(self, elementID)
  if (elementID == 1) then -- the border
    glBeginEnd(GL_QUADS, function(tele)
      local V, N, TC, TN, BN = GetDrawFuncs(tele.verts, tele.norms, tele.txcds,
                                            tele.tangs, tele.binms)
      N(1); TN(1); BN(1); -- west faces
      TC(2); V(2);  TC(3); V(3);  TC(4); V(4);  TC(1); V(1);
      TC(1); V(1);  TC(4); V(4);  TC(5); V(5);  TC(8); V(8);
      TC(8); V(8);  TC(5); V(5);  TC(6); V(6);  TC(7); V(7);
      N(2); TN(2); BN(2); -- east faces
      TC(2); V(15); TC(3); V(14); TC(4); V(13); TC(1); V(16);
      TC(1); V(16); TC(4); V(13); TC(5); V(12); TC(8); V(9);
      TC(8); V(9);  TC(5); V(12); TC(6); V(11); TC(7); V(10);
      N(3); TN(3); BN(3); -- south faces
      TC(1); V(10);  TC(1); V(2); TC(1); V(1); TC(1); V(9);
      TC(1); V(14);  TC(1); V(6); TC(1); V(5); TC(1); V(13);
      N(4); TN(4); BN(4); -- north faces
      TC(1); V(3);  TC(1); V(11); TC(1); V(12); TC(1); V(4);
      TC(1); V(7);  TC(1); V(15); TC(1); V(16); TC(1); V(8);
      N(5); TN(5); BN(5); -- bottom faces
      TC(1); V(2);  TC(1); V(10); TC(1); V(11); TC(1); V(3);
      TC(1); V(4);  TC(1); V(12); TC(1); V(13); TC(1); V(5);
      TC(1); V(6);  TC(1); V(14); TC(1); V(15); TC(1); V(7);
      N(6); TN(6); BN(6); -- top face
      TC(1); V(1);  TC(1); V(8); TC(1); V(16); TC(1); V(9);
    end, self)
  elseif (elementID == 2) then -- links
    glBeginEnd(GL_QUADS, function(tele)
      local V, N, TC, TN, BN = GetDrawFuncs(tele.verts, tele.norms, tele.txcds,
                                            tele.tangs, tele.binms)
      N(1); TN(1); BN(1); -- west link
      TC(1); V(24); TC(1); V(23);  TC(1); V(22);  TC(1); V(21);
      N(1); TN(1); BN(1); -- east link
      TC(1); V(17); TC(1); V(18);  TC(1); V(19);  TC(1); V(20);
    end, self)
  else
    error('DrawWallElement called with elementID = ' .. tostring(elementID))
  end
end


local function DrawWallElement(self, elementID)
  if (elementID ~= 1) then
    error('DrawWallElement called with elementID = ' .. tostring(elementID))
  end
  glBeginEnd(GL_QUADS, function(tele)
    local V, N, TC, TN, BN = GetDrawFuncs(tele.verts, tele.norms, tele.txcds,
                                          tele.tangs, tele.binms)
    N(1); TN(1); BN(1);
    TC(1); V(1); TC(2); V(2); TC(3); V(3); TC(4); V(4);
  end, self)
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function TransformPoint(px, py, pz, tx, ty, tz, cosv, sinv)
  local rx = ((px * cosv) - (py * sinv)) + tx
  local ry = ((px * sinv) + (py * cosv)) + ty
  local rz = pz + tz
  return rx, ry, rz
end


local function TransformNormal(nx, ny, nz, cosv, sinv)
  return
    ((nx * cosv) - (ny * sinv)),
    ((nx * sinv) + (ny * cosv)),
    nz
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CreateBoxGeom(boxID, txcdSideScale, txcdRoofScale)
  if (not boxID) then
    return
  end

  local px, py, pz = bzGetObstaclePosition(boxID)
  local sx, sy, sz = bzGetObstacleSize(boxID)
  local rot        = bzGetObstacleRotation(boxID)

  local cosv = math_cos(rot)  
  local sinv = math_sin(rot)

  local verts = {
    { TransformPoint(-sx, -sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( sx, -sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( sx,  sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx,  sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx, -sy, sz, px, py, pz, cosv, sinv) },
    { TransformPoint( sx, -sy, sz, px, py, pz, cosv, sinv) },
    { TransformPoint( sx,  sy, sz, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx,  sy, sz, px, py, pz, cosv, sinv) },
  }
  local norms = {
    { TransformNormal(-1,  0,  0, cosv, sinv) },
    { TransformNormal( 1,  0,  0, cosv, sinv) },
    { TransformNormal( 0, -1,  0, cosv, sinv) },
    { TransformNormal( 0,  1,  0, cosv, sinv) },
    { TransformNormal( 0,  0, -1, cosv, sinv) },
    { TransformNormal( 0,  0,  1, cosv, sinv) },
  }
  local tangs = {
    norms[3], norms[4], norms[2], norms[1], norms[2], norms[2]
  }
  local binms = {
    norms[6], norms[6], norms[6], norms[6], norms[3], norms[4]
  }

  local ws, ss, ds
  local wt, st, dt
  if (txcdSideScale < 0) then
    ws = -txcdSideScale
    ss = -txcdSideScale
    wt = -txcdSideScale
    st = -txcdSideScale
  else
    ws = sy * 2 * txcdSideScale
    ss = sx * 2 * txcdSideScale
    wt = sz * txcdSideScale
    st = sz * txcdSideScale
  end
  if (txcdRoofScale < 0) then
    ds = -txcdRoofScale
    dt = -txcdRoofScale
  else
    ds = sx * 2 * txcdSideScale
    dt = sy * 2 * txcdSideScale
  end
  local txcds = {
    { 0, 0 }, { ws, 0 }, { ws, wt }, { 0, wt }, -- west
    { 0, 0 }, { ss, 0 }, { ss, st }, { 0, st }, -- south
    { 0, 0 }, { ds, 0 }, { ds, dt }, { 0, dt }, -- down
  }

  return {
    verts = verts,
    norms = norms,
    txcds = txcds,
    tangs = tangs,
    binms = binms,
  }
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CreateBox(boxID)
  local box = CreateBoxGeom(boxID, 0.2, 0.1)
  box.id = boxID
  box.DrawElement = DrawBoxElement
  return box, {
    boxSide = { 1 },
    boxRoof = { 2 },
  }
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CreateBase(baseID)
  local teamNum = bzGetObstacleTeam(baseID)
  local roofName = 'baseRoof' .. teamNum
  local sideName = 'baseSide' .. teamNum
  local base = CreateBoxGeom(baseID, 1, -1)
  base.id = baseID
  base.DrawElement = DrawBoxElement
  return base, {
    [sideName] = { 1 },
    [roofName] = { 2 },
  }
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CreatePyr(pyrID) -- FIXME -- norm / tang / binm / txcds ?
  if (not pyrID) then
    return
  end

  local px, py, pz = bzGetObstaclePosition(pyrID)
  local sx, sy, sz = bzGetObstacleSize(pyrID)
  local rot        = bzGetObstacleRotation(pyrID)
  local flipz      = bzGetObstacleFlipZ(pyrID) -- FIXME -- not handled

  if (flipz) then
    sx, sz = -sx, -sz
  end

  local cosv = math_cos(rot)  
  local sinv = math_sin(rot)

  local verts = {
    { TransformPoint(-sx, -sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( sx, -sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( sx,  sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx,  sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(  0,   0, sz, px, py, pz, cosv, sinv) },
  }
  local norms = {
    { normalize(TransformNormal(-sz,   0, sx, cosv, sinv)) },
    { normalize(TransformNormal( sz,   0, sx, cosv, sinv)) },
    { normalize(TransformNormal(  0, -sz, sy, cosv, sinv)) },
    { normalize(TransformNormal(  0,  sz, sy, cosv, sinv)) },
    { TransformNormal( 0,  0, -1, cosv, sinv) },
  }
  local tangs = {
    { TransformNormal( 0, -1,  0, cosv, sinv) },
    { TransformNormal( 0,  1,  0, cosv, sinv) },
    { TransformNormal( 1,  0,  0, cosv, sinv) },
    { TransformNormal(-1,  0,  0, cosv, sinv) },
    { TransformNormal( 1,  0,  0, cosv, sinv) },
  }
  local binms = {
    { normalize(TransformNormal( sx,   0, sz, cosv, sinv)) },
    { normalize(TransformNormal(-sx,   0, sz, cosv, sinv)) },
    { normalize(TransformNormal(  0,  sy, sz, cosv, sinv)) },
    { normalize(TransformNormal(  0, -sy, sz, cosv, sinv)) },
    { TransformNormal( 0, -1,  0, cosv, sinv) },
  }

  local bs = 0.1 -- bottom txcd scale
  local ss = 0.1 -- side   txcd scale
  local txcds = {
    { 0, 0 },
    { sy * ss, 0 }, { sy * ss * 0.5, ss * length(sx, sz) }, -- x-axis sides
    { sx * ss, 0 }, { sx * ss * 0.5, ss * length(sy, sz) }, -- y-axis sides
    { 0, 0 }, { sx * bs, 0 }, { sx * bs, sy * bs}, { 0, sy * bs } -- base
  }

  if (flipz) then
    for i = 1, #verts do
      local v = verts[i]
      v[3] = v[3] - sz
    end
    local NM = norms
    NM[1], NM[2], NM[3], NM[4] = NM[2], NM[1], NM[4], NM[3]
    local TG = tangs
    TG[1], TG[2], TG[3], TG[4] = TG[2], TG[1], TG[4], TG[3]
    local BN = binms
    BN[1], BN[2], BN[3], BN[4] = BN[2], BN[1], BN[4], BN[3]
  end

  local pyr = {
    id = pyrID,
    DrawElement = DrawPyrElement,
    verts = verts,
    norms = norms,
    txcds = txcds,
    tangs = tangs,
    binms = binms,
  }

  return pyr, {
    pyrSide = { 1 },
    pyrRoof = { 2 },
  }
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CreateTele(teleID) -- FIXME -- txcds
  if (not teleID) then
    return
  end

  local px, py, pz = bzGetObstaclePosition(teleID)
  local sx, sy, sz = bzGetObstacleSize(teleID)
  local rot        = bzGetObstacleRotation(teleID)
  local border     = bzGetObstacleBorder(teleID)

  local cosv = math_cos(rot)  
  local sinv = math_sin(rot)

  local b  = border
  local hb = b * 0.5
  local yo = sy    
  local yi = sy - b
  local zh = sz
  local zl = sz - b
  local verts = {
    -- border
    { TransformPoint( hb, -yo, zh, px, py, pz, cosv, sinv) },
    { TransformPoint( hb, -yo,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( hb, -yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( hb, -yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint( hb,  yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint( hb,  yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( hb,  yo,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( hb,  yo, zh, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb, -yo, zh, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb, -yo,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb, -yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb, -yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb,  yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb,  yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb,  yo,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-hb,  yo, zh, px, py, pz, cosv, sinv) },
    -- links
    { TransformPoint( sx, -yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint( sx, -yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( sx,  yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint( sx,  yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx, -yi, zl, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx, -yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx,  yi,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(-sx,  yi, zl, px, py, pz, cosv, sinv) },
  }
  local norms = {
    { TransformNormal(-1,  0,  0, cosv, sinv) },
    { TransformNormal( 1,  0,  0, cosv, sinv) },
    { TransformNormal( 0, -1,  0, cosv, sinv) },
    { TransformNormal( 0,  1,  0, cosv, sinv) },
    { TransformNormal( 0,  0, -1, cosv, sinv) },
    { TransformNormal( 0,  0,  1, cosv, sinv) },
  }
  local tangs = {
    norms[3], norms[4], norms[2], norms[1], norms[2], norms[2]
  }
  local binms = {
    norms[6], norms[6], norms[6], norms[6], norms[3], norms[4]
  }

  local s  = 1.0 -- texture scale
  local syo = yo * s
  local syi = yi * s
  local szh = zh * s
  local szl = zl * s
  local sb  = b  * s

  local txcds = {
    { -syo, szh }, { -syo, 0 }, { -syi, 0 }, { -syi, szl },
    {  syi, szl }, {  syi, 0 }, {  syo, 0 }, {  syo, szh },
  }

  local tele = {
    id = teleID,
    DrawElement = DrawTeleElement,
    verts = verts,
    norms = norms,
    txcds = txcds,
    tangs = tangs,
    binms = binms,
  }

  return tele, {
    teleBorder =  { 1 },
    teleLink   =  { 2 }
  }
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CreateWall(wallID)
  if (not wallID) then
    return
  end

  local px, py, pz = bzGetObstaclePosition(wallID)
  local sx, sy, sz = bzGetObstacleSize(wallID)
  local rot        = bzGetObstacleRotation(wallID)

  local cosv = math_cos(rot)  
  local sinv = math_sin(rot)

  local verts = {
    { TransformPoint(0, -sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(0,  sy,  0, px, py, pz, cosv, sinv) },
    { TransformPoint(0,  sy, sz, px, py, pz, cosv, sinv) },
    { TransformPoint(0, -sy, sz, px, py, pz, cosv, sinv) },
  }
  local norms = {
    { TransformNormal( 1,  0,  0, cosv, sinv) },
  }
  local tangs = {
    { TransformNormal( 0,  1,  0, cosv, sinv) },
  }
  local binms = {
    { TransformNormal( 0,  0,  1, cosv, sinv) },
  }
  local txcds = {
    { 0, 0 }, { sy, 0 }, { sy, sz}, { 0, sz }
  }

  local wall = {
    id = wallID,
    DrawElement = DrawWallElement,
    verts = verts,
    norms = norms,
    txcds = txcds,
    tangs = tangs,
    binms = binms,
  }

  return wall, { wall = { 1 } }
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local createFuncs = {
  [BZ.OBSTACLE.box]  = CreateBox,
  [BZ.OBSTACLE.base] = CreateBase,
  [BZ.OBSTACLE.pyr]  = CreatePyr,
  [BZ.OBSTACLE.tele] = CreateTele,
  [BZ.OBSTACLE.wall] = CreateWall,
}


local function CreatePrimitive(obsID)
  local type = bzGetObstacleType(obsID)
  local func = createFuncs[type]
  if (func) then
    return func(obsID)
  else
    print('unhandled type: ' .. tostring(BZ.OBSTACLE[type]))
  end
  return nil
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

return CreatePrimitive

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

  