--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--
--  file:     primitives.lua
--  author:   Dave Rodgers (aka: trepan)
--  date:     Jan 07, 2008
--  license:  LGPL 2.1
--  desc:     lua world rendering
--
--  TODO:
--  - implement the rest of the BzMaterial properties
--  - improve the groud segmentation
--  - luaGfxQual control parameter
--  - setup material sets for different qualities
--  - split out a material.lua
--  - overdraw reduction via gl.TexGen
--
--
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local maxLights = 8

local CreatePrimitive = VFS.Include('primitives.lua')

local webroot = 'http://trepan.bzflag.bz/'
local function WebTex(name, notAnRGBS)
  if (notAnRGBS) then
    return bz.GetCacheFilePath(webroot .. name)
  else
    return bz.GetCacheFilePath(webroot .. 'rgbs_nmdp/' .. name)
  end
end


-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

local function bzPrintf(fmt, ...)
  bz.Print(fmt:format(...))
end


-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

local defaultShaders  = {}
local perPixelShaders = {}
local parallaxShaders = {}
local reliefShaders   = {}

local shaderMap = {
  default  = defaultShaders,
  perpixel = perPixelShaders,
  parallax = parallaxShaders,
  relief   = reliefShaders,
}
local defShaderType = 'perpixel'



local function SetShaderType(type)
  if (not shaderMap[type]) then
    return false
  end
  defShaderType = type
  return true
end


SetShaderType('parallax')
SetShaderType('relief')
SetShaderType('perpixel')
  


-- export
function CreateShaders()
  local startTime = bz.GetTimer()
  for type, shaderTable in pairs(shaderMap) do
    for maxLights = -1, 7 do
      local vertCode = VFS.ReadFile(type .. '.vert')
      local fragCode = VFS.ReadFile(type .. '.frag')

      if (maxLights >= 0) then
        fragCode = '#define LAST_LIGHT ' .. maxLights .. '\n' .. fragCode
      end

      shader = gl.CreateShader({
        vertex   = vertCode,
        fragment = fragCode,
        uniform = {
          depthFactor   =  0.04, -- for relief
          parallaxBias  =  0.03, -- for parallax
          parallaxScale = -0.03, -- for parallax

        },
        uniformInt = {
          bzTexture0 = 0,
          bzTexture1 = 1,
        }
      })

      if (shader == nil) then
        local msg = string.format(
          'Could not compile the %s shader for %i lights', type, maxLights
        )
        bz.Print(msg)
        print(msg)
        bz.Print(gl.GetShaderLog())
        print(gl.GetShaderLog())
      end

      shaderTable[maxLights] = shader
    end
  end
  local endTime = bz.GetTimer()
  print(bz.DiffTimers(endTime, startTime))
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local wantMyWorld = true
local haveMyWorld = false

local triCount
local quadCount
local polyCount
local totalTris


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--
--  Materials
--

local mats = {}

local matBins = {}

local matLists = {}


-- ordered lists of references to 'mats'
local matOpaque = {} -- FIXME -- not used yet

local matTranslucent = {}


local function CreateDefaultMaterial()
  return {
		name = 'luadefault',
		order = 0,

		ambient   = { 0.2, 0.2, 0.2, 1.0 },
		diffuse   = { 1.0, 1.0, 1.0, 1.0 },
		emission  = { 0.0, 0.0, 0.0, 1.0 },
		specular  = { 0.0, 0.0, 0.0, 1.0 },
		shininess = 16.0,

		dyncol      = false,
		occluder    = false,
		groupAlpha  = false,
		noLighting  = false,
		noRadar     = false,
		noShadow    = false,
		noCulling   = false,
		noSorting   = false,
		alphaThresh = 0.0,

		textures = {},

		shaders = {},
  }
end


local defMaterial = CreateDefaultMaterial()
defMaterial.visible = true


local lastMaterial = defMaterial


local function CompareMaterials(matA, matB)
  -- order
  if (matA.order < matB.order) then return true end
  if (matB.order < matA.order) then return false end

  -- shader
  local shaderA = matA.shaders[1]
  local shaderB = matA.shaders[1]
  if (shaderA and (not shaderB)) then return true end
  if (shaderB and (not shaderA)) then return false end
  if (shaderA and (shaderA ~= shaderB)) then
    if (shaderA < shaderB) then return true end
    if (shaderB < shaderA) then return false end
  end

  -- textures
  local texSetA = matA.textures
  local texSetB = matB.textures
  local texCountA = #texSetA
  local texCountB = #texSetB
  if (texCountA < texCountB) then return true end
  if (texCountB < texCountA) then return false end
  for t = 1, texCountA do
    -- texture name
    if (texSetA[t].name < texSetB[t].name) then return true end
    if (texSetB[t].name < texSetA[t].name) then return false end
    -- texture matrix
    if ((not texSetA[t].texmat) and texSetB[t].texmat) then return true end
    if ((not texSetB[t].texmat) and texSetA[t].texmat) then return false end
  end

  -- lighting
  if (matA.noLighting and (not matB.noLighting)) then return true end
  if (matB.noLighting and (not matA.noLighting)) then return false end

  -- dynamic color
  if ((not matA.dyncol) and matB.dyncol) then return true end
  if ((not matB.dyncol) and matA.dyncol) then return false end

  return false
end


local function SetupMaterialOrder(mats)
  matOpaque = {}
  matTranslucent = {}
  local mo = 0
  local mt = 0
  for id, mat in pairs(mats) do
    if (mat.diffuse[3] >= 1.0) then
      mo = mo + 1
      matOpaque[mo] = mat
    else
      mt = mt + 1
      matTranslucent[mt] = mat
    end
  end
  table.sort(matOpaque, CompareMaterials)
--  table.print(matOpaque, function(...) print('OPAQUE:', ...) end) -- FIXME
end


local function Shaderize(mats)
  for _, mat in pairs(mats) do
    mat.textures[1] = { name = WebTex('boxtop_rgbs.png') }
    mat.textures[2] = { name = WebTex('boxtop_nmdp.png') }
  end
end


local texMap = {}
do
  local exts   = { '', '.png' }
  local types  = { 'wall', 'top' }
  local colors = { 'red', 'green', 'blue', 'purple' }
  for _, e in ipairs(exts) do
    for _, t in ipairs(types) do
      for _, c in ipairs(colors) do
        texMap[c .. '_' .. 'base' .. t .. e] =
          'skins/'.. c .. '/base' .. t .. '.png'
      end
    end
  end
end


local function SetupMaterials()

  mats = {} -- clear the current list

  -- engine mesh materials
  for matID = 1, bz.GetMaterialCount() do
    local mat = {}
    mats[matID] = mat

		mat.name = bz.GetMaterialName(matID)
		mat.order = bz.GetMaterialOrder(matID)

		mat.ambient   = { bz.GetMaterialAmbient(matID)  }
		mat.diffuse   = { bz.GetMaterialDiffuse(matID)  }
		mat.emission  = { bz.GetMaterialEmission(matID) }
		mat.specular  = { bz.GetMaterialSpecular(matID) }
		mat.shininess = bz.GetMaterialShininess(matID)

		mat.dyncol      = bz.GetMaterialDynCol(matID)
		mat.occluder    = bz.GetMaterialOccluder(matID)
		mat.groupAlpha  = bz.GetMaterialGroupAlpha(matID)
		mat.noLighting  = bz.GetMaterialNoLighting(matID)
		mat.noRadar     = bz.GetMaterialNoRadar(matID)
		mat.noShadow    = bz.GetMaterialNoShadow(matID)
		mat.noCulling   = bz.GetMaterialNoCulling(matID)
		mat.noSorting   = bz.GetMaterialNoSorting(matID)
		mat.alphaThresh = bz.GetMaterialAlphaThresh(matID)

		mat.textures = {}
		for t = 1, bz.GetMaterialTextureCount(matID) do
		  mat.textures[t] = bz.GetMaterialTextureInfo(matID, t)
    end

    mat.shaders = {}		
		for s = 1, bz.GetMaterialShaderCount(matID) do
		  mat.shaders[s] = bz.GetMaterialShader(matID, s)
    end
  end

  -- materials for primitives
  local types = {
    'ground',
    'wall',
    'boxRoof', 'boxSide',
    'pyrRoof', 'pyrSide',
    'baseRoof1', 'baseSide1',
    'baseRoof2', 'baseSide2',
    'baseRoof3', 'baseSide3',
    'baseRoof4', 'baseSide4',
    'teleBorder', 'teleLink',
  }
  for _, type in ipairs(types) do
    mats[type] = CreateDefaultMaterial()
    mats[type].name = type
    mats[type].shininess = 32.0
    mats[type].specular = { 1, 1, 1, 1 }
  end

  -- standard
  mats.ground.textures[1]  = { name = 'std_ground.png' }
  mats.wall.textures[1]    = { name = 'wall.png' }
  mats.boxRoof.textures[1] = { name = 'roof.png' }
  mats.boxSide.textures[1] = { name = 'boxwall.png' }
  mats.pyrRoof.textures[1] = { name = 'pyrwall.png' }
  mats.pyrSide.textures[1] = { name = 'pyrwall.png' }
  mats.teleBorder.textures[1] = { name = 'caution.png' }
  mats.teleLink.textures[1]   = { name = 'mesh.png'    }
  mats.teleLink.diffuse[4]   = 0.75
  mats.baseRoof1.textures[1] = { name = 'skins/red/basetop.png'  }
  mats.baseSide1.textures[1] = { name = 'skins/red/basewall.png' }
  mats.baseRoof2.textures[1] = { name = 'skins/green/basetop.png'  }
  mats.baseSide2.textures[1] = { name = 'skins/green/basewall.png' }
  mats.baseRoof3.textures[1] = { name = 'skins/blue/basetop.png'  }
  mats.baseSide3.textures[1] = { name = 'skins/blue/basewall.png' }
  mats.baseRoof4.textures[1] = { name = 'skins/purple/basetop.png'  }
  mats.baseSide4.textures[1] = { name = 'skins/purple/basewall.png' }

  mats.pyrSide.diffuse[4] = 0.5

  mats.teleLink.shaders[1]  = 'perpixel'
  mats.teleLink.textures[1] = { name = WebTex('lua_logo.png', true) }
  mats.teleLink.diffuse[4]   = 0.25
---[[
  mats.ground.shaders[1]  = 'relief'
  mats.ground.textures[1] = { name = WebTex('ground_rgbs.png') }
  mats.ground.textures[2] = { name = WebTex('ground_nmdp.png') }
--]]
---[[
  mats.wall.shaders[1]  = 'relief'
  mats.wall.textures[1] = { name = WebTex('wall_rgbs.png') }
  mats.wall.textures[2] = { name = WebTex('wall_nmdp.png') }
--  mats.wall.diffuse     = { 1.0, 0.5, 0,0 }
--]]
---[[
  mats.boxRoof.shaders[1] = 'relief'
  mats.boxRoof.textures[1] = { name = WebTex('boxtop_rgbs.png') }
  mats.boxRoof.textures[2] = { name = WebTex('boxtop_nmdp.png') }
  mats.boxSide.shaders[1] = 'relief'
  mats.boxSide.textures[1] = { name = WebTex('boxwall_rgbs.png') }
  mats.boxSide.textures[2] = { name = WebTex('boxwall_nmdp.png') }
--]]
---[[
  mats.pyrRoof.shaders[1] = 'relief'
  mats.pyrRoof.textures[1] = { name = WebTex('pyrwall_rgbs.png') }
  mats.pyrRoof.textures[2] = { name = WebTex('pyrwall_nmdp.png') }
  mats.pyrSide.shaders[1] = 'relief'
  mats.pyrSide.textures[1] = { name = WebTex('pyrwall_rgbs.png') }
  mats.pyrSide.textures[2] = { name = WebTex('pyrwall_nmdp.png') }
--]]


---[[
  mats.baseRoof1.noLighting = true
  mats.baseSide1.noLighting = true
  mats.baseRoof2.noLighting = true
  mats.baseSide2.noLighting = true
  mats.baseRoof3.noLighting = true
  mats.baseSide3.noLighting = true
  mats.baseRoof4.noLighting = true
  mats.baseSide4.noLighting = true
--]]
  mats.baseRoof1.shaders[1] = 'perpixel'
  mats.baseSide1.shaders[1] = 'perpixel'
  mats.baseRoof2.shaders[1] = 'perpixel'
  mats.baseSide2.shaders[1] = 'perpixel'
  mats.baseRoof3.shaders[1] = 'perpixel'
  mats.baseSide3.shaders[1] = 'perpixel'
  mats.baseRoof4.shaders[1] = 'perpixel'
  mats.baseSide4.shaders[1] = 'perpixel'

  for tag, mat in pairs(mats) do
    mat.tag = tag
    mat.visible = (mat.diffuse[4] > 0.0)
    if (mat.shininess == 0) then
      mat.shininess = 1
    end
    for _, tex in ipairs(mat.textures) do
--      tex.name = texMap[tex.name] or tex.name
    end
--    mat.textures[1] = { name = 'lua_logo.png' }
--    mat.textures[2] = { name = 'lua_logo.png' }
  end

--  Shaderize(mats)

  SetupMaterialOrder(mats)
end
		    

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------


local activeLightLast = -1
local activeLightCount = 0


local function ExecuteMaterial(newMat)

--  print('material: '..lastMaterial.name..' -> '..newMat.name)
  
  if (newMat == lastMaterial) then
    return true
  elseif (not newMat.visible) then
    print('... skipped, invisible')
    return false
  end

--  table.print({ [newMat.name] = newMat })

  local oldMat = lastMaterial
  lastMaterial = newMat

  gl.Blending(newMat.diffuse[4] < 1.0)
  if (newMat.diffuse[4] <  1.0) then
--    print('blend for ' .. newMat.name .. ' @ ' .. newMat.diffuse[4]) -- FIXME
  end

  if (newMat.noLighting) then
    if (not oldMat.noLighting) then
      gl.Lighting(false)
    end
    gl.Color(newMat.diffuse)
  else
    if (oldMat.noLighting) then
      gl.Lighting(true)
    end
    gl.Ambient(newMat.ambient)
    gl.Diffuse(newMat.diffuse)
    gl.Emission(newMat.emission)
    gl.Specular(newMat.specular)
    gl.Shininess(newMat.shininess)
  end
    gl.Ambient(newMat.ambient)
    gl.Diffuse(newMat.diffuse)
    gl.Emission(newMat.emission)
    gl.Specular(newMat.specular)
    gl.Shininess(newMat.shininess)

  local oldTexCount = #oldMat.textures
  local newTexCount = #newMat.textures
--  print('oldTexCount = ' .. oldTexCount)
--  print('newTexCount = ' .. newTexCount)
  local maxTex = math.max(oldTexCount, newTexCount)
  for t = 1, maxTex do
    local oldTex = oldMat.textures[t]
    local newTex = newMat.textures[t]
    local oldTexName = oldTex and oldTex.name
    local newTexName = newTex and newTex.name
--    print('texUnit' .. t - 1, oldTexName, newTexName)
    if (oldTexName ~= newTexName) then
--      print('texture' .. (t - 1), newTexName or false)
      gl.Texture(t - 1, newTexName or false) -- FIXME
    end
  end

  local oldShader = oldMat.shaders[1]
  local newShader = newMat.shaders[1]
  if (oldShader ~= newShader) then
    local shaderSet = shaderMap[newShader]
    local shader = shaderSet and shaderSet[activeLightLast]
    gl.UseShader(shader or false)
  end

  local alphaThresh = newMat.alphaThresh
  if (alphaThresh and (alphaThresh > 0.0)) then
    gl.AlphaTest(GL.GREATER, alphaThresh)
  else
    gl.AlphaTest(false)
  end

  return true
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function vec2sub(v1, v2)
  return (v1[1] - v2[1]),
         (v1[2] - v2[2])
end
local function vec3sub(v1, v2)
  return (v1[1] - v2[1]),
         (v1[2] - v2[2]),
         (v1[3] - v2[3])
end


local normalize = math.vector.normalize
local cross     = math.vector.cross


local function CalcTangentBinorms(plane, verts, norms, txcds)
  local Nx, Ny, Nz
  if (not norms) then
    Nx, Ny, Nz = normalize(plane[1], plane[2], plane[3])
  else
    Nx, Ny, Nz = normalize(norms[1][1], norms[1][2], norms[1][3])
  end

  local Ax, Ay, Az = normalize(vec3sub(verts[2], verts[1]))
  local Bx, By, Bz = cross(Nx, Ny, Nz, Ax, Ay, Az)
  local Dx, Dy = normalize(vec2sub(txcds[2], txcds[1]))

  local Sx, Sy, Sz = ((Ax * Dx) - (Bx * Dy)),
                     ((Ay * Dx) - (By * Dy)),
                     ((Az * Dx) - (Bz * Dy))

  local Tx, Ty, Tz = ((Ax * Dy) + (Bx * Dx)),
                     ((Ay * Dy) + (By * Dx)),
                     ((Az * Dy) + (Bz * Dx))

  Sx, Sy, Sz = normalize(Sx, Sy, Sz)

  Tx, Ty, Tz = normalize(Tx, Ty, Tz)

  local tangents = {}
  local binorms = {}
  for i = 1, #verts do
    tangents[i] = { Sx, Sy, Sz }
    binorms[i]  = { Tx, Ty, Tz }
  end
  return tangents, binorms
end


local function DrawMeshFaces(meshID, faceList)
  -- collect the 
  local triList,  tcount  = {}, 0
  local quadList, qcount = {}, 0
  local polyList, pcount = {}, 0
  for _, faceID in ipairs(faceList) do
    local elements = bz.GetFaceElementCount(meshID, faceID)
    if (elements == 3) then
      tcount = tcount + 1
      triList[tcount] = faceID
    elseif (elements == 4) then
      qcount = qcount + 1
      quadList[qcount] = faceID
    else
      pcount = pcount + 1
      polyList[pcount] = faceID
    end
  end
  triCount  = triCount  + tcount
  quadCount = quadCount + qcount
  polyCount = polyCount + pcount

  local glTypeMap = {
    [GL.TRIANGLES] = { 3, triList  },
    [GL.QUADS]     = { 4, quadList },
    [GL.POLYGON]   = { 5, polyList },
  }
  for glType, shapeSpec in pairs(glTypeMap) do
    local shapeVerts = shapeSpec[1]
    local shapeList  = shapeSpec[2]
    for _, faceID in ipairs(shapeList) do
      gl.BeginEnd(glType, function(meshID, faceID)
        local verts = bz.GetFaceVerts(meshID, faceID)
        local norms = bz.GetFaceNorms(meshID, faceID)
        local txcds = bz.GetFaceTxcds(meshID, faceID) --, false (do not use auto-txcds)
        local plane = bz.GetFacePlane(meshID, faceID)
        if (not norms) then
          gl.Normal(plane)
        end
        local tang, binorm = CalcTangentBinorms(plane, verts, norms, txcds)

        for v = 1, #verts do
          if (norms) then gl.Normal(norms[v])   end
          if (txcds) then
            gl.TexCoord(txcds[v])
            gl.MultiTexCoord(1, txcds[v])
            gl.MultiTexCoord(2, tang[v])
            gl.MultiTexCoord(3, binorm[v])
          end
          gl.Vertex(verts[v])
        end

        totalTris = totalTris + (#verts - 2)

      end, meshID, faceID)
    end
  end
end


local function DrawFaces(list)
  for binTag, elementList in pairs(list) do
    if (isnumber(binTag)) then
      -- each element is a face
      DrawMeshFaces(binTag, elementList)
    elseif (istable(binTag)) then
      for e = 1, #elementList do
        binTag:DrawElement(elementList[e])
      end
    end
  end
end


local function GetMatBin(tag)
  local mat = mats[tag]
  if (not mat) then
    error('not a material tag: ' .. tostring(tag))
  end
  local matBin = matBins[mat]
  if (not matBin) then
    matBin = {}
    matBins[mat] = matBin
  end
  return matBin
end


local function AddMeshFaces(meshID)
  local faceCount = bz.GetObstacleFaceCount(meshID)
  if (faceCount) then
    for faceID = 1, faceCount do
      local matID = bz.GetFaceMaterial(meshID, faceID)
      local matBin = GetMatBin(matID)
      local meshBin = matBin[meshID]
      if (not meshBin) then
        meshBin = {}
        matBin[meshID] = meshBin
      end
      meshBin[#meshBin + 1] = faceID
    end
  end
end


local function AddPrimFaces(prim, matFaces)
  for matTag, faceList in pairs(matFaces) do
    local matBin = GetMatBin(matTag)
    matBin[prim] = faceList
  end
end    


-- export
function CreateDLists()
  local startTime = bz.GetTime(1000)
  SetupMaterials()
  local endTime = bz.GetTime(1000)
  bzPrintf('materials setup in %.3f seconds', endTime - startTime)

  triCount  = 0
  quadCount = 0
  polyCount = 0
  totalTris = 0

  matBins = {}

  matLists = {}

  local startTime = bz.GetTimer()
  local obsList = bz.GetObstacleList()
  for _, obsID in ipairs(obsList) do
    if (bz.HasMeshDrawInfo(obsID)) then
      -- FIXME local x = bz.GetMeshDrawInfo(obsID)
    else
      local type = bz.GetObstacleType(obsID)
      if (type == BZ.OBSTACLE.mesh) then
        AddMeshFaces(obsID)
      else
        local obj, matFaces = CreatePrimitive(obsID)
        if (obj) then
          AddPrimFaces(obj, matFaces)
        end
      end
    end
  end
  local endTime = bz.GetTimer()
  bzPrintf('material sorting took %.3f seconds', bz.DiffTimers(endTime, startTime))

  local startTime = bz.GetTimer()
  for mat, list in pairs(matBins) do
    if (mat and mat.visible) then
      local dlist = gl.CreateList(DrawFaces, list)
      if (dlist) then
        matLists[mat] = dlist
      end
    end
  end
  local endTime = bz.GetTimer()
  bzPrintf('dlists created in %.3f seconds', bz.DiffTimers(endTime, startTime))

--  print('totalTris = ' .. totalTris)

  -- clear the lists to save memory
--  mats = {}
  matBins = {}
  collectgarbage()
end


local function DrawGround()

  local cx, cy, cz = bz.GetCameraPosition()

  local ws = BZDB.GetFloat('_worldSize')
  local s = ws * 0.5
  local ts = 0.25 -- texture scale

  local maxDist = 200
  local x0, x1, x2 = -s, cx, s
  local y0, y1, y2 = -s, cy, s
  local s0, s1, s2 = -s * ts, cx * ts, s * ts  
  local t0, t1, t2 = -s * ts, cy * ts, s * ts  

  
  -- FIXME
--  gl.TexRect(-s, -s, s, s, -s * ts, -s * ts, s * ts, s * ts)
---[[
  gl.Normal(0, 0, 1)
  gl.MultiTexCoord(2, 1, 0, 0)
  gl.MultiTexCoord(3, 0, 1, 0)
  gl.TexRect(x0, y0, x1, y1, s0, t0, s1, t1)
  gl.TexRect(x1, y0, x2, y1, s1, t0, s2, t1)
  gl.TexRect(x1, y1, x2, y2, s1, t1, s2, t2)
  gl.TexRect(x0, y1, x1, y2, s0, t1, s1, t2)
--]]
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function SetupLights()
  activeLightLast = -1
  activeLightCount = 0
  local GL_LIGHT0  = 0x4000
  local GL_AMBIENT = 0x1200
  for light = 0, 7 do
    local on = gl.GetNumber(GL_LIGHT0 + light)
    if (on ~= 0.0) then
      activeLightLast = light
      activeLightCount = activeLightCount + 1
      gl.Light(light, GL_AMBIENT, 0, 0, 0, 1)
    else
      gl.Light(light, GL_AMBIENT, 0, 0, 0, 0)
    end
  end
--FIXME  print(activeLightLast)
end


local function HaveMirror()
  return
    (BZDB.GetString('_mirror') ~= 'none') and
    (BZDB.GetBool('userMirror'))          and
    (not bz.GetGfxBlock('mirror'))
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function CallRawLists()
  for mat, dlist in pairs(matLists) do
    gl.CallList(dlist)
  end
end


local function CallMatLists()
  for mat, dlist in pairs(matLists) do
    if (ExecuteMaterial(mat)) then
      gl.CallList(dlist)
    end
  end
end


local function DrawWorld()
  if (not next(matLists)) then
    return
  end

  SetupLights()

  local noMirror = not HaveMirror()

  gl.DepthTest(GL.LEQUAL)
  gl.DepthMask(true)

  gl.Culling(GL.BACK)
  gl.Lighting(true) -- default state

  ExecuteMaterial(defMaterial)

  if (1 > 0) then -- zpass
    gl.Clear(GL.DEPTH_BUFFER_BIT)
    gl.ColorMask(false, false, false, false)
    if (noMirror) then
      CallRawLists()
      DrawGround()
    end
    gl.ColorMask(true, true, true, true)
  end

  CallMatLists()

  if (noMirror) then
    ExecuteMaterial(mats.ground);
    DrawGround()
  end

  ExecuteMaterial(defMaterial)

  gl.Culling(false)
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function Enable()
  CreateDLists()
  if (next(matLists)) then
    CreateShaders()
  end
  Script.SetCallIn('DrawWorldStart', function()
    local success, err = xpcall(DrawWorld, traceback)
    if (not success) then
      print(err)
      Script.SetCallIn('DrawWorldStart', nil)
    end
  end)
end


function Disable()
  for name, shaderSet in pairs(shaderMap) do
    for key, shader in pairs(shaderSet) do
      gl.DeleteShader(shader)
      key = nil
    end
  end
  matLists = {}
  collectgarbage()
  Script.SetCallIn('DrawWorldStart', nil)
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function GotGfxBlock(type, id)
  if ((type ~= BZ.GFXBLOCKTYPE.global) or (id ~= BZ.GFXBLOCKID.obstacles)) then
    return
  end
--  if (haveMyWorld or (not wantMyWorld)) then
--    return
--  end
  haveMyWorld = true
  Enable()
end

Script.SetCallIn('GotGfxBlock', GotGfxBlock)



local function LostGfxBlock(type, id)
  if ((type ~= BZ.GFXBLOCKTYPE.global) or (id ~= BZ.GFXBLOCKID.obstacles)) then
    return
  end
  haveMyWorld = false
  Disable()
end

Script.SetCallIn('LostGfxBlock', LostGfxBlock)


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

Script.SetCallIn('RecvCommand', function(msg)
  local s, e, cmd, arg = msg:find('(%S*)%s*(%S*)')
  if (cmd ~= 'remap') then
    return
  end
  if (arg == 'false') then
    Disable()
    wantMyWorld = false  
    haveMyWorld = false
    bz.SetGfxBlock('obstacles', false)
    return
  end
  local newType = SetShaderType(arg)
  if (newType or (arg == 'true')) then
    wantMyWorld = true
    Disable()
    if (bz.SetGfxBlock('obstacles', true, true)) then
      GotGfxBlock(BZ.GFXBLOCKTYPE.global, BZ.GFXBLOCKID.obstacles)
    end
  end
end)


Script.SetCallIn('ServerParted', function()
  matLists = {}
  collectgarbage()
end)


Script.SetCallIn('ServerJoined', function()
  matLists = {}
  CreateDLists()
  collectgarbage()
end)


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

if (wantMyWorld) then
  if (bz.SetGfxBlock('obstacles', true, true)) then
    haveMyWorld = false
    GotGfxBlock(BZ.GFXBLOCKTYPE.global, BZ.GFXBLOCKID.obstacles)
  end
end



if (-1 > 0) then
  p = { 0, 0, 1 }
  v = { { 0, 0, 0 }, { 1, 0, 0 } }
  n = { { 0, 0, 1 }, { 0, 0, 1 } }
  t = { { 0, 0 }, { 2, 1 } }
  table.print({CalcTangentBinorms(p, v, n, t)})
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function WorldGLReload()
  if (not haveMyWorld) then
    return
  end
  local startTime = bz.GetTimer()
  CreateDLists()
  if (next(matLists)) then
    CreateShaders()
  end
  local endTime = bz.GetTimer()
  local secs = bz.DiffTimers(endTime, startTime)
  bz.Print('WorldGLReload took ' .. secs .. ' seconds')
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

